/* manual.css */

body { 
  background: #fffbd7; 
  color: #000000; 
  padding: 15pt;
  font-family: Helvetica,Verdana,Arial,sans-serif;
  text-align: justify;
}

h1 { font-size: 200%; }
h2 { font-size: 160%; }
h3 { font-size: 160%; }
h4 { font-size: 130%; }
h5 { font-size: 100%; }

.pcenter {
  text-align: center;
}

.pleft {
  text-align: left;
}

.pright {
  text-align: right;
}

div.func {
  background-color: #eeeac6; 
  margin: 0pt 0pt; 
}

table {
  margin-left: auto;
  margin-right: auto;
}

td, th {
  border-collapse: collapse;
  border-style: none;
} 

table.func { 
  padding: 0pt 1ex; 
  margin-left: 1ex;
  margin-right: 1ex;
  background-color: transparent;
  /* line-height: 1.1; */ 
  width: 100%; 
}

table.example {
  background-color: #efefd7; 
  border-style: none; 
  border-width: 0pt; 
  padding: 0px;  
  width: 100%
}
table.example td {
  border-style: none;
  border-width: 0pt;
  padding: 0ex 1ex;
}

table.chlink {
  background-color: #dddddd; 
  border-style: none; 
  border-width: thin;  
  text-align: center;
  /* width: 100%;  */
}

table.chlink td {
  border-style: groove;
  border-width: thin;
}

table.GAPDocTable {
  border-collapse: collapse;
  border-style: none;
  border-color: black;
}

table.GAPDocTable  td, table.GAPDocTable  th {
  padding: 3pt;
  border-width: thin;
  border-style: solid;
  border-color: #555555;
}

caption.GAPDocTable {
  caption-side: bottom;
  width: 70%;
  margin-top: 1em;
  margin-left: auto;
  margin-right: auto;
}

td.tdleft {
  text-align: left;
}

td.tdright {
  text-align: right;
}

td.tdcenter {
  text-align: center;
}

div.chlink {
  text-align: center;
}
td.chlink1 {
  background-color: #cccccc;
}

pre { 
  font-family: "Courier New",Courier,monospace; 
  font-size: 100%; 
  color:#284d49; 
}

tt,code { 
  font-family: "Courier New",Courier,monospace; 
  font-size: 100%; 
  color: #284d49; }

var {
  color: #cf0000;
}

pre.normal {
  color: #000000;
}

code.func {
  color: #006500;
}
code.keyw {
  color: #ff0000;
}
code.file {
  color: #8e4510;
}

p.foot {
  font-size: 60%;
  font-style: normal;
}

a:link { 
  color: #00008e; 
  text-decoration: none; 
}
a:visited { 
  color: #00008e; 
  text-decoration: none; 
}
a:active { 
  color: #000000; 
  text-decoration: none; 
}

